/****************************************************************
	LLPAM Emulator
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University
	Last modified by Mutsunori Banbara on Jaly. 27, 2000
	(banbara@cs32.scitec.kobe-u.ac.jp)
****************************************************************/
/*
	This program is based on the following text book.
	@Book{AitKaci91:wam,
	  author = "Hassan A{\"\i}t-Kaci",
	  title = "Warren's Abstract Machine",
	  publisher = "The MIT Press",
	  year = 1991
	}
*/

#include <stdio.h>

#define EXTERN
#include "llpam.h"
#include "resource.h"
#include "debug.h"
#include "instr.h"

#define AM_INSTR INLINE void

/****************************************************************
	Subroutines
****************************************************************/
void am_abort(int abortcode)
{
    abort_code = abortcode;
    printf("*** LLPAM error %d ", abort_code);
    printf("at %ld = store + %ld\n", P, PTR2INX(P));
    am_list_instr(P);
    halt = YES;
    longjmp(am_jmpbuf, 1);
}


INLINE void am_check()
{
    LONG used;
    if (prolog_flag.max_usage) {
	used = sizeof(CELL) * (code_bot - code_top);	/* code */
	used += sizeof(CELL) * (H - heap_top);		/* heap */
	used += sizeof(CELL) * (AM_STACK_BOT - stack_top);	/* stack */
	used += sizeof(CELL) * (TR - trail_top);		/* trail */
	used += sizeof(sym_struct) * symbol_bot;		/* symbol */
	used += sizeof(res_struct) * R;			/* resource */
	used += sizeof(CELL) * (2*HASHSIZE);		/* hash, hash2 */
	if (used > max_usage) {
	    max_usage = used;
	}
    }

    if (H + MARGIN >= stack_top) {
	printf(" HEAP area (%ld) overflow\n", HEAPSIZE);
	am_abort(102);
    }
    if (E + MARGIN >= trail_top || B + MARGIN >= trail_top) {
	printf("STACK area (%ld) overflow\n", STACKSIZE);
	am_abort(103);
    }
    if (TR + MARGIN >= pdl_bot) {
	printf("TRAIL area (%ld) overflow\n", TRAILSIZE);
	am_abort(104);
    }
    if (R + RESMARGIN >= RESSIZE) {
	printf("RES area (%ld) overflow\n", RESSIZE);
	am_abort(105);
    }
}


void am_unwind_trail(STOREPTR a1, STOREPTR a2)
{
    STOREPTR p;
    CELL d, x;
    RESINX k, m;
    LONG n;
    void am_dump_cell();
    BOOLEAN b;

    for (p = a2 - 1; p >= a1; )  {
	d = STORE(p);
	switch (TAG(d)) {
	  case TagREF:
	    STORE(REF2PTR(d)) = MAKEREF(REF2PTR(d));
	    p--;
	    break;
	  case TagCON:
	    RLIST = CDR(RLIST);
	    p--;
	    break;
	  case TagINT:
	    n = INT2LONG(d);
	    /* undo changelevel and changedeadline */
	    m = RES2RESINX(STORE(--p));
	    b = TOP2BOOL(STORE(--p));
#ifdef DEBUG
	    if ((trace & TRACE_AM) != 0) { 
		if (b) {
		    printf("Undo Level ");
		    am_print_term(RES(k).head);
		    printf(" : ");
		    printf("%ld", RES(k).level);
		    printf(" -> ");
		    printf("%ld\n", n);
		} else {
		    printf("Undo Deadline ");
		    am_print_term(RES(k).head);
		    printf(" : ");
		    printf("%ld", RES(k).deadline);
		    printf(" -> ");
		    printf("%ld\n", n);		    
		}
	    }
#endif 
	    if (b) {
	        for (k = RES(m).s1; k < RES(m).s2; k++)
		    RES(k).level = n;
	    } else {
	        for (k = RES(m).s1; k < RES(m).s2; k++)
		    RES(k).deadline = n;
	    }
	    p--;
	    break;
	  case TagLIS:
	    x = STORE(LIS2PTR(d));
	    STORE(LIS2PTR(d)) = CDR(x);
	    p--;
	    break;
	  case TagRES:
	    /* undo end_imp and end_exp_imp */
	    m = RES2RESINX(d);
	    k = RES2RESINX(STORE(--p));
	    for ( ; k < m; k++)
		RES(k).out_of_scope = NO;
	    p--;
	    break;
	  default:
	    printf("Internal error in am_unwind_trail\n");
	    am_abort(200);
	}
    }
}


void am_tidy_trail()
{
}


int am_occurs_check(CELL x, CELL y)
{
    STOREPTR p;
    CELL d;
    int t, i, n;

    AM_DEREF(d, t, y);
    /* x should be REF, LIS, or STR */
    if (x == d)
	return YES;
    switch (t) {
      case TagREF:
      case TagINT:
      case TagCON:
      case TagRES:
	  return NO;
      case TagLIS:
	  p = LIS2PTR(d);
	  return am_occurs_check(x, STORE(p))
	      || am_occurs_check(x, STORE(p+1));
      case TagSTR:
	  p = STR2PTR(d);
	  n = SYMBOL(CON2SYM(STORE(p))).arity;
	  for (i = 1; i <= n; i++) {
	      if (am_occurs_check(x, STORE(p+i)))
		  return YES;
	  }
	  return NO;
    }
    return NO;
}


void am_unify(CELL a1, CELL a2)
{
    STOREPTR pdl, p1, p2;
    CELL d1, d2;
    int t1, t2, i, n;

    pdl = pdl_bot;
    *--pdl = a1;
    *--pdl = a2;
    while (!fail && pdl < pdl_bot) {
	AM_DEREF(d2, t2, *pdl++);
	AM_DEREF(d1, t1, *pdl++);
	if (d1 == d2) {
	    continue;
	}
	if (t1 == TagREF || t2 == TagREF) {
	    if (t1 == TagREF && (t2 != TagREF || d2 < d1)) {
		if (AM_OCCURS_CHECK(d1, d2)) {
		    fail = YES;
		    continue;
		}
		AM_ASSIGN(d1, d2);
	    } else {
		if (AM_OCCURS_CHECK(d2, d1)) {
		    fail = YES;
		    continue;
		}
		AM_ASSIGN(d2, d1);
	    }
	    continue;
	}
	if (t1 != t2) {
	    fail = YES;
	    continue;
	}
	switch (t1) {
	  case TagINT:
	  case TagCON:
	  case TagRES:
	    fail = (d1 != d2);
	    break;
	  case TagLIS:
	    p1 = LIS2PTR(d1);
	    p2 = LIS2PTR(d2);
	    *--pdl = STORE(p1);
	    *--pdl = STORE(p2);
	    *--pdl = STORE(p1+1);
	    *--pdl = STORE(p2+1);
	    if (pdl - MARGIN <= pdl_top) {
		printf("PDL area (%ld) overflow\n", PDLSIZE);
		am_abort(107);
	    }
	    break;
	  case TagSTR:
	    p1 = STR2PTR(d1);
	    p2 = STR2PTR(d2);
	    if (STORE(p1) != STORE(p2)) {
		fail = YES;
		break;
	    }
	    n = SYMBOL(CON2SYM(STORE(p1))).arity;
	    for (i = 1; i <= n; i++) {
		*--pdl = STORE(p1+i);
		*--pdl = STORE(p2+i);
	    }
	    if (pdl - MARGIN <= pdl_top) {
		printf("PDL area (%ld) overflow\n", PDLSIZE);
		am_abort(106);
	    }
	    break;
	}
    }
}


INLINE CELL am_indirect(CELL x)
{
    STOREPTR p;
    int i;

    switch (TAG(x)) {
    case TagCON:
	num_of_args = 0;
	break;
    case TagSTR:
	p = STR2PTR(x);
	x = STORE(p);
	num_of_args = SYMBOL(CON2SYM(x)).arity;
	for (i = 1; i <= num_of_args; i++)
	    A(i) = STORE(p+i);
	break;
    default:
	x = UNDEF;
	break;
    }
    return(x);
}


void am_exec_indirect(CELL goal)
{
    SYMBOLINX pred;

    if (goal == SYM_TRUE) {
	P = CP;
	return;
    }
    if ((predicate = am_indirect(goal)) == UNDEF) {
	printf("Indirect call error\n");
	am_abort(2);
    }
    pred = CON2SYM(predicate);
    B0 = B;
    if (SYMBOL(pred).codeaddr == PRED_UNDEF) {
	if (prolog_flag.unknown == 0 && TAG(SYMBOL(pred).res3) != TagLIS) {
	    printf("Undefined predicate ");
	    am_print_functor_arity(predicate);
	    printf("\n");
	}
	AM_FAIL;
	return;
	/* am_abort(1); */
    }
    if (TAG(SYMBOL(pred).res) != TagLIS) {
	R1 = R2 = SYM_NIL;
    } else {
	lookup_hash(pred);
    }
    P = SYMBOL(pred).codeaddr;
    am_check();
}

/****************************************************************
	Special Instructions
****************************************************************/
AM_INSTR am_undefined()
{
    printf("Illegal instruction\n");
    am_abort(300);
}


AM_INSTR am_halt()
{
    halt = YES;
    longjmp(am_jmpbuf, 1);
}


AM_INSTR am_fail()
{
    AM_FAIL;
}


AM_INSTR am_trace()
{
    trace = CARG1;
    P++;
    longjmp(am_jmpbuf, 1);
}


AM_INSTR am_nop()
{
    P++;
}


AM_INSTR am_deref_a()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG1));
    A(CARG1) = x;
    P++;
}


AM_INSTR am_deref_y()
{
    CELL x;
    int t;

    AM_DEREF(x, t, Y(CARG1));
    Y(CARG1) = x;
    P++;
}


/****************************************************************
	DEBUG Instructions
****************************************************************/
AM_INSTR am_debug_dmp_a()
{
    printf("A(%d) = ", CARG1);
    am_dump_term(A(CARG1));
    P++;
}


AM_INSTR am_debug_dmp_y()
{
    printf("Y(%d) = ", CARG1);
    am_dump_term(Y(CARG1));
    P++;
}


AM_INSTR am_debug_prn_a()
{
    printf("A(%d) = ", CARG1);
    am_print_term(A(CARG1));
    printf("\n");
    P++;
}


AM_INSTR am_debug_prn_y()
{
    printf("Y(%d) = ", CARG1);
    am_print_term(Y(CARG1));
    printf("\n");
    P++;
}


AM_INSTR am_debug_dmp_e()
{
    int i, n;

    n = STORE(E+2);
    printf("E = %ld\n", PTR2INX(E));
    printf("  CE    : %5ld\n", PTR2INX((STOREPTR)STORE(E)));
    printf("  CP    : %5ld\n", PTR2INX((STOREPTR)STORE(E+1)));
    printf("  n     : %5ld\n", (LONG)n);
    for (i = 1; i <= n; i++) {
	printf("  Y(%3d): ", i);
	am_dump_cell(STORE(E+i+2));
	printf("\n");
    }
    P++;
}


AM_INSTR am_debug_dmp_b()
{
    int i, n;

    n = STORE(B);
    printf("B = %ld\n", PTR2INX(B));
    for (i = 1; i <= n; i++)  {
        printf("  A(%3d): ", i);
	am_dump_cell(STORE(B+i));
	printf("\n");
    }
    printf("  CE    : %5ld\n", PTR2INX((STOREPTR)STORE(B+n+1)));
    printf("  CP    : %5ld\n", PTR2INX((STOREPTR)STORE(B+n+2)));
    printf("  B     : %5ld\n", PTR2INX((STOREPTR)STORE(B+n+3)));
    printf("  BP    : %5ld\n", PTR2INX((STOREPTR)STORE(B+n+4)));
    printf("  TR    : %5ld\n", PTR2INX((STOREPTR)STORE(B+n+5)));
    printf("  H     : %5ld\n", PTR2INX((STOREPTR)STORE(B+n+6)));
    printf("  B0    : %5ld\n", PTR2INX((STOREPTR)STORE(B+n+7)));
    printf("  R     : %5ld\n", (LONG)STORE(B+n+8));
    printf("  L     : %5ld\n", (LONG)STORE(B+n+9));
    printf("  U     : %5ld\n", (LONG)STORE(B+n+10));
    P++;
}


AM_INSTR am_debug_dmp_s(){
    SYMBOLINX i;

    for (i=0; i<symbol_bot; i++) {
	printf("No. %4ld [", (LONG)i);
	printf("%s/",  SYMBOL(i).print_name);
	printf("%d ]\n", SYMBOL(i).arity);
	printf("         hash = %5ld\n", SYMBOL(i).hash_value);
	printf("         addr = %5ld\n", (LONG)SYMBOL(i).codeaddr);
	printf("         res  = ");  am_print_term(SYMBOL(i).res);
	printf("\n");
	printf("         res2 = ");  am_print_term(SYMBOL(i).res2); 
	printf("\n");
	printf("         res3 = ");  am_print_term(SYMBOL(i).res3); 
	printf("\n");
	printf("         res4 = ");  am_print_term(SYMBOL(i).res4); 
	printf("\n");
    }
    P++;
}

/****************************************************************
	PUT Instructions
****************************************************************/
/* put_var Xi, Aj */
AM_INSTR am_put_var_a()
{
    A(CARG1) = A(CARG2) = STORE(H) = MAKEREF(H);
    H++;
    P++;
}


/* put_var Yi, Aj */
AM_INSTR am_put_var_y()
{
    STOREPTR p;

    p = E + CARG1 + 2;	/* Pointer to Yi */
    A(CARG2) = STORE(p) = MAKEREF(p);
    P++;
}


/* put_val Xi, Aj */
AM_INSTR am_put_val_a()
{
    A(CARG2) = A(CARG1);
    P++;
}


/* put_val Yi, Aj */
AM_INSTR am_put_val_y()
{
    A(CARG2) = Y(CARG1);
    P++;
}


/* put_unsafe_val Yi, Aj */
AM_INSTR am_put_unsafe_val()
{
    CELL x;
    int t;

    AM_DEREF(x, t, Y(CARG1));
    if (t != TagREF || REF2PTR(x) < E) {
        A(CARG2) = x;
    } else {
        A(CARG2) = STORE(H) = MAKEREF(H);
        AM_ASSIGN(x, STORE(H));
        H++;
    }
    P++;
}

/* put_str f/n, Ai */
AM_INSTR am_put_str()
{
    A(CARG1) = MAKESTR(H);
    STORE(H++) = CARGW(1);
    P += 2;
}


/* put_list Ai */
AM_INSTR am_put_list()
{
    A(CARG1) = MAKELIS(H);
    P++;
}


/* put_con c, Ai */
AM_INSTR am_put_con()
{
    A(CARG1) = CARGW(1);
    P += 2;
}


/* put_int n, Ai */
AM_INSTR am_put_int()
{
    A(CARG1) = CARGW(1);
    P += 2;
}


/* put_ground Term, Ai */
AM_INSTR am_put_ground()
{
    A(CARG1) = CARGW(1);
    P += 2;
}


/****************************************************************
	GET Instructions
****************************************************************/
/* get_var Xi, Aj */
AM_INSTR am_get_var_a()
{
    A(CARG1) = A(CARG2);
    P++;
}


/* get_var Yi, Aj */
AM_INSTR am_get_var_y()
{
    Y(CARG1) = A(CARG2);
    P++;
}


/* get_val Xi, Aj */
AM_INSTR am_get_val_a()
{
    am_unify(A(CARG1), A(CARG2));
    if (fail) {
	am_fail();
	return;
    }
    P++;
}


/* get_val Yi, Aj */
AM_INSTR am_get_val_y()
{
    am_unify(Y(CARG1), A(CARG2));
    if (fail) {
	am_fail();
	return;
    }
    P++;
}


/* get_str f/n, Ai */
AM_INSTR am_get_str()
{
    CELL x, f;
    int t;

    f = CARGW(1);
    AM_DEREF(x, t, A(CARG1));
    switch (t) {
      case TagREF:
	last_get = MAKESTR(H);
	AM_ASSIGN(x, last_get);
	STORE(H++) = f;
	readmode = NO;
	break;
      case TagSTR:
	S = STR2PTR(x);
	if (STORE(S) != f) {
	    AM_FAIL;
	    return;
        }
	S++;
	readmode = YES;
	break;
      default:
	AM_FAIL;
	return;
    }
    P += 2;
}


/* get_list Ai */
AM_INSTR am_get_list()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG1));
    switch (t) {
      case TagREF:
	last_get = MAKELIS(H);
	AM_ASSIGN(x, last_get);
	readmode = NO;
	break;
      case TagLIS:
	S = LIS2PTR(x);
	readmode = YES;
	break;
      default:
	AM_FAIL;
	return;
    }
    P++;
}


/* get_con c, Ai */
AM_INSTR am_get_con()
{
    CELL x, f;
    int t;

    f = CARGW(1);
    AM_DEREF(x, t, A(CARG1));
    switch (t) {
      case TagREF:
	AM_ASSIGN(x, f);
	break;
      case TagCON:
	if (x != f) {
	    AM_FAIL;
	    return;
	}
	break;
      default:
	AM_FAIL;
	return;
    }
    P += 2;
}


/* get_int n, Ai */
AM_INSTR am_get_int()
{
    CELL x, d;
    int t;

    d = CARGW(1);
    AM_DEREF(x, t, A(CARG1));
    switch (t) {
      case TagREF:
	AM_ASSIGN(x, d);
	break;
      case TagINT:
	if (x != d) {
	    AM_FAIL;
	    return;
	}
	break;
      default:
	AM_FAIL;
	return;
    }
    P += 2;
}


/* get_ground Term, Ai */
AM_INSTR am_get_ground()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG1));
    if (t == TagREF) {
	AM_ASSIGN(x, CARGW(1));
    } else {
	am_unify(x, CARGW(1));
	if (fail) {
	    am_fail();
	    return;
	}
    }
    P += 2;
}


/****************************************************************
	SET Instructions
****************************************************************/
/* set_var Ai */
AM_INSTR am_set_var_a()
{
    A(CARG1) = STORE(H) = MAKEREF(H);
    H++;
    P++;
}


/* set_var Yi */
AM_INSTR am_set_var_y()
{
    Y(CARG1) = STORE(H) = MAKEREF(H);
    H++;
    P++;
}


/* set_val Ai */
AM_INSTR am_set_val_a()
{
    STORE(H++) = A(CARG1);
    P++;
}


/* set_val Yi */
AM_INSTR am_set_val_y()
{
    STORE(H++) = Y(CARG1);
    P++;
}


/* set_local_val Ai */
AM_INSTR am_set_local_val_a()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG1));
    if (t != TagREF || REF2PTR(x) < H) {
        STORE(H++) = x;
    } else {
        STORE(H) = MAKEREF(H);
        AM_ASSIGN(x, STORE(H));
        H++;
    }
    P++;
}


/* set_local_val Yi */
AM_INSTR am_set_local_val_y()
{
    CELL x;
    int t;

    AM_DEREF(x, t, Y(CARG1));
    if (t != TagREF || REF2PTR(x) < H) {
        STORE(H++) = x;
    } else {
        STORE(H) = MAKEREF(H);
        AM_ASSIGN(x, STORE(H));
        H++;
    }
    P++;
}


/* set_con c */
AM_INSTR am_set_con()
{
    STORE(H++) = CARGW(1);
    P += 2;
}


/* set_int n */
AM_INSTR am_set_int()
{
    STORE(H++) = CARGW(1);
    P += 2;
}


/* set_ground Term */
AM_INSTR am_set_ground()
{
    STORE(H++) = CARGW(1);
    P += 2;
}


/* set_void n */
AM_INSTR am_set_void()
{
    int i;

    for (i = 0; i < CARG1; i++) {
        STORE(H) = MAKEREF(H);
	H++;
    }
    P++;
}


/****************************************************************
	UNIFY Instructions
****************************************************************/
/* unify_var Ai */
AM_INSTR am_unify_var_a_r()
{
    A(CARG1) = STORE(S++);
    P++;
}

AM_INSTR am_unify_var_a_w()
{
    A(CARG1) = STORE(H) = MAKEREF(H);
    H++;
    P++;
}


/* unify_var Yi */
AM_INSTR am_unify_var_y_r()
{
    Y(CARG1) = STORE(S++);
    P++;
}

AM_INSTR am_unify_var_y_w()
{
    Y(CARG1) = STORE(H) = MAKEREF(H);
    H++;
    P++;
}


/* unify_val Ai */
AM_INSTR am_unify_val_a_r()
{
    am_unify(A(CARG1), STORE(S++));
    if (fail) {
	am_fail();
	return;
    }
    P++;
}

AM_INSTR am_unify_val_a_w()
{
    if (AM_OCCURS_CHECK(last_get, A(CARG1))) {
	AM_FAIL;
	return;
    }
    STORE(H++) = A(CARG1);
    P++;
}


/* unify_val Yi */
AM_INSTR am_unify_val_y_r()
{
    am_unify(Y(CARG1), STORE(S++));
    if (fail) {
	am_fail();
	return;
    }
    P++;
}

AM_INSTR am_unify_val_y_w()
{
    if (AM_OCCURS_CHECK(last_get, Y(CARG1))) {
	AM_FAIL;
	return;
    }
    STORE(H++) = Y(CARG1);
    P++;
}


/* unify_local_val Ai */
AM_INSTR am_unify_local_val_a_r()
{
    am_unify(A(CARG1), STORE(S++));
    if (fail) {
	am_fail();
	return;
    }
    P++;
}

AM_INSTR am_unify_local_val_a_w()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG1));
    if (AM_OCCURS_CHECK(last_get, x)) {
        AM_FAIL;
        return;
    }
    if (t != TagREF || REF2PTR(x) < H) {
	STORE(H++) = x;
    } else {
	STORE(H) = MAKEREF(H);
	AM_ASSIGN(x, STORE(H));
	H++;
    }
    P++;
}


/* unify_local_val Yi */
AM_INSTR am_unify_local_val_y_r()
{
    am_unify(Y(CARG1), STORE(S++));
    if (fail) {
	am_fail();
	return;
    }
    P++;
}

AM_INSTR am_unify_local_val_y_w()
{
    CELL x;
    int t;

    AM_DEREF(x, t, Y(CARG1));
    if (t != TagREF || REF2PTR(x) < H) {
	STORE(H++) = x;
    } else {
	STORE(H) = MAKEREF(H);
	AM_ASSIGN(x, STORE(H));
	H++;
    }
    P++;
}


/* unify_con c */
AM_INSTR am_unify_con_r()
{
    CELL x;
    int t;

    AM_DEREF(x, t, STORE(S++));
    if (t == TagREF) {
	AM_ASSIGN(x, CARGW(1));
    }
    else if (x != CARGW(1)) {
	AM_FAIL;
	return;
    }
    P += 2;
}

AM_INSTR am_unify_con_w()
{
    STORE(H++) = CARGW(1);
    P += 2;
}


/* unify_int n */
AM_INSTR am_unify_int_r()
{
    CELL x;
    int t;

    AM_DEREF(x, t, STORE(S++));
    if (t == TagREF) {
	AM_ASSIGN(x, CARGW(1));
    }
    else if (x != CARGW(1)) {
	AM_FAIL;
	return;
    }
    P += 2;
}

AM_INSTR am_unify_int_w()
{
    STORE(H++) = CARGW(1);
    P += 2;
}


/* unify_ground Term */
AM_INSTR am_unify_ground_r()
{
    CELL x;
    int t;

    AM_DEREF(x, t, STORE(S++));
    if (t == TagREF) {
	AM_ASSIGN(x, CARGW(1));
    } else {
	am_unify(x, CARGW(1));
	if (fail) {
	    am_fail();
	    return;
	}
    }
    P += 2;
}

AM_INSTR am_unify_ground_w()
{
    STORE(H++) = CARGW(1);
    P += 2;
}


/* unify_void n */
AM_INSTR am_unify_void_r()
{
    S += CARG1;
    P++;
}

AM_INSTR am_unify_void_w()
{
    int i;

    for (i = 0; i < CARG1; i++) {
	STORE(H) = MAKEREF(H);
	H++;
    }
    P++;
}


/****************************************************************
	CONTROL Instructions
****************************************************************/
/* allocate n */
AM_INSTR am_allocate()
{
    STOREPTR newE;

    newE = AM_STACK_BOT;
    STORE(newE) = (CELL)E;
    STORE(newE+1) = (CELL)CP;
    STORE(newE+2) = (CELL)CARG1;
    E = newE;
    P++;
}


/* deallocate */
AM_INSTR am_deallocate()
{
    CP = (STOREPTR)STORE(E+1);
    E = (STOREPTR)STORE(E);
    P++;
}


/* call f/n */
AM_INSTR am_call()
{
    SYMBOLINX pred;

    predicate = CARGW(1);
    pred = CON2SYM(predicate);
    num_of_args = SYMBOL(pred).arity;
    CP = P + 2;
    B0 = B;
    if (SYMBOL(pred).codeaddr == PRED_UNDEF) {
	if (prolog_flag.unknown == 0) {
	    printf("Undefined predicate ");
	    am_print_functor_arity(predicate);
	    printf("\n");
	    am_abort(1);
	}
	AM_FAIL;
	return;
    }
    if (TAG(SYMBOL(pred).res) != TagLIS) {
	R1 = R2 = SYM_NIL;
    } else {
	lookup_hash(pred);
    }
    P = SYMBOL(pred).codeaddr;
    am_check();
}


AM_INSTR am_call_trace()
{
    SYMBOLINX pred;
    CELL goal;
    int i;

    if ((trace & TRACE_CALL) == 0) {
	am_call();
	return;
    }
    predicate = CARGW(1);
    pred = CON2SYM(predicate);
    num_of_args = SYMBOL(pred).arity;
    goal = MAKESTR(H);
    STORE(H++) = predicate;
    for (i = 1; i <= num_of_args; i++)
      STORE(H++) = A(i);
    CP = P + 2;
    am_trace_goal(goal);
}


/* execute f/n */
AM_INSTR am_execute()
{
    SYMBOLINX pred;

    predicate = CARGW(1);
    pred = CON2SYM(predicate);
    num_of_args = SYMBOL(pred).arity;
    B0 = B;
    if (SYMBOL(pred).codeaddr == PRED_UNDEF) {
	if (prolog_flag.unknown == 0) {
	    printf("Undefined predicate ");
	    am_print_functor_arity(predicate);
	    printf("\n");
	    am_abort(1);
	}
	AM_FAIL;
	return;
    }
    if (TAG(SYMBOL(pred).res) != TagLIS) {
	R1 = R2 = SYM_NIL;
    } else {
	lookup_hash(pred);
    }
    P = SYMBOL(pred).codeaddr;
    am_check();
}


AM_INSTR am_execute_trace()
{
    SYMBOLINX pred;
    CELL goal;
    int i;

    if ((trace & TRACE_CALL) == 0) {
	am_execute();
	return;
    }
    predicate = CARGW(1);
    pred = CON2SYM(predicate);
    num_of_args = SYMBOL(pred).arity;
    goal = MAKESTR(H);
    STORE(H++) = predicate;
    for (i = 1; i <= num_of_args; i++)
      STORE(H++) = A(i);
    am_trace_goal(goal);
}


/* execute_indirect Ai */
AM_INSTR am_execute_indirect()
{
    SYMBOLINX pred;

    /*
       A(CARG1) should be dereferenced beforehand
    */
    am_exec_indirect(A(CARG1));
}


AM_INSTR am_execute_indirect_trace()
{
    CELL goal;

    if ((trace & TRACE_CALL) == 0) {
	am_execute_indirect();
	return;
    }
    goal = A(CARG1);
    am_trace_goal(goal);
}


/* execute_indirect_notrace Ai */
AM_INSTR am_execute_indirect_notrace()
{
    /*
       A(CARG1) should be dereferenced beforehand
    */
    am_exec_indirect(A(CARG1));
}


/* Perhaps not needed */
/* execute_prog_indirect Ai */
/* resources are not be executed */
AM_INSTR am_execute_prog_indirect()
{
    SYMBOLINX pred;

    /*
       A(CARG1) should be dereferenced beforehand
    */
    if (A(CARG1) == SYM_TRUE) {
	P = CP;
	return;
    }
    if ((predicate = am_indirect(A(CARG1))) == UNDEF) {
	printf("Indirect call error\n");
	am_abort(2);
    }
    pred = CON2SYM(predicate);
    B0 = B;
    P = SYMBOL(pred).codeaddr;
    if (P == PRED_UNDEF) {
	AM_FAIL;
	return;
    }
    am_check();
}


/* proceed */
AM_INSTR am_proceed()
{
    P = CP;
}


/****************************************************************
	CHOICE Instructions
****************************************************************/
/* try_me_else L */
AM_INSTR am_try_me_else()
{
    STOREPTR label, newB, p;
    int i, n;

    label = INX2PTR(CARGW(1));
    newB = AM_STACK_BOT;
    n = STORE(newB) = num_of_args;
    for (i = 1; i <= n; i++)
      STORE(newB+i) = A(i);
    STORE(newB+n+1) = (CELL)E;
    STORE(newB+n+2) = (CELL)CP;
    STORE(newB+n+3) = (CELL)B;
    STORE(newB+n+4) = (CELL)label;
    STORE(newB+n+5) = (CELL)TR;
    STORE(newB+n+6) = (CELL)H;
    STORE(newB+n+7) = (CELL)B0;
    STORE(newB+n+8) = R;
    STORE(newB+n+9) = L;
    STORE(newB+n+10) = U;
    STORE(newB+n+11) = (CELL)T;
    B = newB;
    HB = H;
    P += 2;
}


/* retry_me_else L */
AM_INSTR am_retry_me_else()
{
    STOREPTR label;
    int i, n;

    label = INX2PTR(CARGW(1));
    n = num_of_args = STORE(B);
    for (i = 1; i <= n; i++)
      A(i) = STORE(B+i);
    E = (STOREPTR)STORE(B+n+1);
    CP = (STOREPTR)STORE(B+n+2);
    STORE(B+n+4) = (CELL)label;
    am_unwind_trail((STOREPTR)STORE(B+n+5), TR);
    TR = (STOREPTR)STORE(B+n+5);
    HB = H = (STOREPTR)STORE(B+n+6);
    R = STORE(B+n+8);
    L = STORE(B+n+9);
    U = STORE(B+n+10);
    T = (BOOLEAN)STORE(B+n+11);
    P += 2;
}


/* trust_me */
AM_INSTR am_trust_me()
{
    int i, n;

    n = num_of_args = STORE(B);
    for (i = 1; i <= n; i++)
      A(i) = STORE(B+i);
    E = (STOREPTR)STORE(B+n+1);
    CP = (STOREPTR)STORE(B+n+2);
    am_unwind_trail((STOREPTR)STORE(B+n+5), TR);
    TR = (STOREPTR)STORE(B+n+5);
    H = (STOREPTR)STORE(B+n+6);
    R = STORE(B+n+8);
    L = STORE(B+n+9);
    U = STORE(B+n+10);
    T = (BOOLEAN)STORE(B+n+11);
    B = (STOREPTR)STORE(B+n+3);
    HB = (STOREPTR)STORE(B+STORE(B)+6);
    P++;
}


/* try L */
AM_INSTR am_try()
{
    STOREPTR label, newB;
    int i, n;

    label = INX2PTR(CARGW(1));
    newB = AM_STACK_BOT;
    n = STORE(newB) = num_of_args;
    for (i = 1; i <= n; i++)
      STORE(newB+i) = A(i);
    STORE(newB+n+1) = (CELL)E;
    STORE(newB+n+2) = (CELL)CP;
    STORE(newB+n+3) = (CELL)B;
    STORE(newB+n+4) = (CELL)(P + 2);
    STORE(newB+n+5) = (CELL)TR;
    STORE(newB+n+6) = (CELL)H;
    STORE(newB+n+7) = (CELL)B0;
    STORE(newB+n+8) = R;
    STORE(newB+n+9) = L;
    STORE(newB+n+10) = U;
    STORE(newB+n+11) = (CELL)T;
    B = newB;
    HB = H;
    P = label;
}


/* retry L */
AM_INSTR am_retry()
{
    STOREPTR label;
    int i, n;

    label = INX2PTR(CARGW(1));
    n = num_of_args = STORE(B);
    for (i = 1; i <= n; i++)
      A(i) = STORE(B+i);
    E = (STOREPTR)STORE(B+n+1);
    CP = (STOREPTR)STORE(B+n+2);
    STORE(B+n+4) = (CELL)(P + 2);
    am_unwind_trail((STOREPTR)STORE(B+n+5), TR);
    TR = (STOREPTR)STORE(B+n+5);
    HB = H = (STOREPTR)STORE(B+n+6);
    R = STORE(B+n+8);
    L = STORE(B+n+9);
    U = STORE(B+n+10);
    T = (BOOLEAN)STORE(B+n+11);
    P = label;
}


/* trust L */
AM_INSTR am_trust()
{
    STOREPTR label;
    int i, n;

    label = INX2PTR(CARGW(1));
    n = num_of_args = STORE(B);
    for (i = 1; i <= n; i++)
      A(i) = STORE(B+i);
    E = (STOREPTR)STORE(B+n+1);
    CP = (STOREPTR)STORE(B+n+2);
    am_unwind_trail((STOREPTR)STORE(B+n+5), TR);
    TR = (STOREPTR)STORE(B+n+5);
    H = (STOREPTR)STORE(B+n+6);
    R = STORE(B+n+8);
    L = STORE(B+n+9);
    U = STORE(B+n+10);
    T = (BOOLEAN)STORE(B+n+11);
    B = (STOREPTR)STORE(B+n+3);
    HB = (STOREPTR)STORE(B+STORE(B)+6);
    P = label;
}


/****************************************************************
	INDEXING Instructions
****************************************************************/
/* switch_on_term V, I, C, S, L */    
AM_INSTR am_switch_on_term()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(1));
    switch (t) {
      case TagINT: P = INX2PTR(CARGW(2)); break;
      case TagCON: P = INX2PTR(CARGW(3)); break;
      case TagSTR: P = INX2PTR(CARGW(4)); break;
      case TagLIS: P = INX2PTR(CARGW(5)); break;
      default:     P = INX2PTR(CARGW(1)); break; /* including TagREF */
    }
}


AM_INSTR am_switch_on_hash()
{
    CELL x;
    int t;
    ULONG h;

    AM_DEREF(x, t, A(1));
    switch (t) {
    case TagREF:
	h = 0;
	break;
    case TagINT:
    case TagRES:
	h = x;
	break;
    case TagCON:
	h = SYMBOL(CON2SYM(x)).hash_value;
	break;
    case TagSTR:
	h = SYMBOL(CON2SYM(STORE(STR2PTR(x)))).hash_value;
	break;
    case TagLIS:
	h = SYMBOL(CON2SYM(SYM_DOT)).hash_value;
	break;
    }
    h = h % CARG1;
    P = INX2PTR(CARGW(h + 1));
}


/****************************************************************
	CUT Instructions
****************************************************************/
/* neck_cut */
AM_INSTR am_neck_cut()
{
    if (B > B0) {
	B = B0;
	HB = (STOREPTR)STORE(B+STORE(B)+6);
	am_tidy_trail();
    }
    P++;
}


/* get_level Yi */
AM_INSTR am_get_level()
{
    Y(CARG1) = MAKEINT(PTR2INX(B0));
    P++;
}


/* cut Yi */
AM_INSTR am_cut()
{
    STOREPTR level;

    level = INX2PTR(INT2LONG(Y(CARG1)));
    if (B > level) {
	B = level;
	HB = (STOREPTR)STORE(B+STORE(B)+6);
	am_tidy_trail();
    }
    P++;
}


/****************************************************************
	LLPAM Emulator
****************************************************************/
void am_loop_trace()
{
    if (!readmode)
      goto WRITEMODE;

    for (;;) {
      READMODE:
	if ((trace & TRACE_AM) != 0)
	  am_list_instr(P);
#	define TRACECODE
#	define RMODECODE
#	include "dispatch.c"
    }

    for (;;) {
      WRITEMODE:
	if ((trace & TRACE_AM) != 0)
	  am_list_instr(P);
#	define TRACECODE
#	undef  RMODECODE
#	include "dispatch.c"
    }
}


void am_loop()
{
    if (!readmode)
      goto WRITEMODE;

    for (;;) {
      READMODE:
#	undef  TRACECODE
#	define RMODECODE
#	include "dispatch.c"
	/* debug
	am_list_instr(P);
	printf("(P, CP) = (%ld, %ld)\n ", (long)P, (long)CP);
	*/
    }

    for (;;) {
      WRITEMODE:
#	undef  TRACECODE
#	undef  RMODECODE
#	include "dispatch.c"
	/* debug
	am_list_instr(P);
	printf("(P, CP) = (%ld, %ld)\n ", (long)P, (long)CP);
	*/
    }
}


int am_interpreter()
{
    readmode = YES;
    abort_code = 0;
    halt = NO;
    fail = NO;
    if (setjmp(am_jmpbuf)) {
	if (halt) {
	    return(abort_code);
	}
    }
    if (trace > 0) {
	am_loop_trace();
    } else {
	am_loop();
    }
}


void am_init()
{
    LONG i;

    totalsize = 0;
    totalsize += sizeof(CELL) * CODESIZE;
    totalsize += sizeof(CELL) * HEAPSIZE;
    totalsize += sizeof(CELL) * STACKSIZE;
    totalsize += sizeof(CELL) * TRAILSIZE;
    totalsize += sizeof(CELL) * PDLSIZE;
    totalsize += sizeof(sym_struct) * SYMSIZE;
    totalsize += sizeof(res_struct) * RESSIZE;
    totalsize += sizeof(CELL) * (2*HASHSIZE);
    max_usage = 0;

    /* Allocate STORE */
    store = (STOREPTR)malloc(totalsize);
    if (store == NULL) {
	printf("Can not allocate %d bytes memory\n", totalsize);
	exit(1);
    }
    /* Initialize Area Pointers */
    code_top = code_bot = store;
    heap_top = code_top + CODESIZE;
    stack_top = heap_top + HEAPSIZE;
    trail_top = stack_top + STACKSIZE;
    pdl_top = trail_top + TRAILSIZE;
    pdl_bot = pdl_top + PDLSIZE;
    symbol = (sym_struct*)pdl_bot;
    symbol_bot = (SYMBOLINX)0;
    res = (res_struct*)(symbol + SYMSIZE);
    hash = (CELL*)(res + RESSIZE);
    hash2 = (CELL*)(hash + HASHSIZE);

    /* Special Symbols */
    SYM_NIL    = MAKECON(am_make_symbol("[]",  0));
    SYM_DOT    = MAKECON(am_make_symbol(".",   2));
    SYM_PLUS1  = MAKECON(am_make_symbol("+",   1));
    SYM_PLUS   = MAKECON(am_make_symbol("+",   2));
    SYM_MINUS1 = MAKECON(am_make_symbol("-",   1));
    SYM_MINUS  = MAKECON(am_make_symbol("-",   2));
    SYM_TIMES  = MAKECON(am_make_symbol("*",   2));
    SYM_DIVIDE = MAKECON(am_make_symbol("/",   2));
    SYM_QUOT   = MAKECON(am_make_symbol("//",  2));
    SYM_MOD    = MAKECON(am_make_symbol("mod", 2));
    SYM_SHIFTL = MAKECON(am_make_symbol("<<",  2));
    SYM_SHIFTR = MAKECON(am_make_symbol(">>",  2));
    SYM_NEG    = MAKECON(am_make_symbol("\\",  1));
    SYM_AND    = MAKECON(am_make_symbol("/\\", 2));
    SYM_OR     = MAKECON(am_make_symbol("\\/", 2));
    SYM_XOR    = MAKECON(am_make_symbol("#",   2));
    SYM_ABS    = MAKECON(am_make_symbol("abs", 1));
    SYM_MAX    = MAKECON(am_make_symbol("max", 2));
    SYM_MIN    = MAKECON(am_make_symbol("min", 2));

    SYM_TRUE   = MAKECON(am_make_symbol("true", 0));
    SYM_TRACE  = MAKECON(am_make_symbol("$trace", 1));
    SYM_RES    = MAKECON(am_make_symbol("$res", 2));

    /* initialize hash and hash 2*/
    for (i = 0; i < HASHSIZE; i++)
	HASH(i) = SYM_NIL;
    for (i = 0; i < HASHSIZE; i++)
	HASH2(i) = SYM_NIL;

    /* Trace level */
    trace = 0;

    /* prolog_flag */
    prolog_flag.unknown = 0;		/* unknown = error */
    prolog_flag.occurs_check = 0;	/* occurs_check = off */
    prolog_flag.max_usage = 0;		/* max_usage = off */

    /* initialize registers and stack */
    S = HB = H = heap_top;
    B0 = B = E = stack_top;
    TR = trail_top;
    R0 = R = (RESINX)0;
    L = (LONG)1;
    U = (LONG)-1;
    RLIST = SYM_NIL;
    R1 = R2 = SYM_NIL; 
    T = FALSE;
    STORE(B) = 0;
    STORE(B+1) = (CELL)E;
    STORE(B+2) = (CELL)CP;
    STORE(B+3) = (CELL)B;
    STORE(B+4) = (CELL)B;
    STORE(B+5) = (CELL)TR;
    STORE(B+6) = (CELL)H;
    STORE(B+7) = (CELL)B0;
    STORE(B+8) = (CELL)R;
    STORE(B+9) = (CELL)L;
    STORE(B+10) = (CELL)U;
    STORE(B+11) = (CELL)T;
}


void am_free()
{
    SYMBOLINX s;

    for (s = 0; s < symbol_bot; s++)
      free(SYMBOL(s).print_name);
    free(store);
}


/*
 Local Variables:
 mode: c
 c-indent-level: 4
 End:
*/
