/****************************************************************
	LLPAM Main
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Eiji Sugiyama
	   Dept. CS, Kobe University
****************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>

#include "version.h"
#define EXTERN extern
#include "llpam.h"
#include "instr.h"

int main(int argc, char *argv[])
{
    char *objname;
    int i, rc;
    long runtime;
    char *entry_name;
    SYMBOLINX entry_pred;
    int syslevel;

    printf("%s\n", version);
    if (argc < 2) {
	am_usage(argc, argv);
	exit(1);
    }
    am_init();
    STORE(code_bot++) = (CELL)HALT;	/* first instruction is HALT */
    P = code_bot;
    trace = 0;
    entry_name = "main";
    syslevel = 0;
    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
	      case 't':
		trace = 1;
		if (isdigit(argv[i][2])) {
		    trace = argv[i][2] - '0';
		    printf("Trace level %d\n", trace);
		}
		break;
	      case 'e':
		if (argv[i][2] != '\0') {
		    entry_name = &argv[i][2];
		}
		break;
	      case 's':
		syslevel = 1;
		break;
	      default:
		am_usage(argc, argv);
		exit(1);
	    }
	    continue;
	}
	objname = argv[i];
	if (am_load(objname, syslevel) != 0) {
	    exit(1);
	}
	P = load_base;
	CP = code_top;
	rc = am_interpreter();
	if (rc != 0) {
	    printf("Error during the initialization of %s\n", objname);
	    exit(1);
	}
    }
    entry_pred = am_make_symbol(entry_name, 0);
    P = SYMBOL(entry_pred).codeaddr;
    CP = code_top;
    if (P == PRED_UNDEF) {
	printf("Predicate %s/0 is undefined.\n", entry_name);
	exit(1);
    }
    if ((trace & TRACE_AM) != 0) {
	printf("***** Execution start from %ld\n", PTR2INX(P));
    }
    runtime0 = usertime();

    rc = am_interpreter();

    runtime = usertime() - runtime0;
    if ((trace & TRACE_AM) != 0) {
	printf("***** Execution stopped at %ld ", PTR2INX(P));
	printf("(%ld msec)\n", runtime);
    }
    am_free();
    return(rc);
}


int am_usage(int argc, char *argv[])
{
    printf("%s\n", copyright);
    printf("%s\n", organization);
    printf("%s\n", www);
    printf("\n");
    printf("Usage: %s [options] {[-s] file.llpo}...\n", argv[0]);
    printf("Options   -tN    : trace level (1--9)\n"); 
    printf("          -eName : entry predicate name\n");
    printf("          -s     : system-level predicates\n");
    printf("\n");
    return 0;
}


int am_load(char *fn, int level)
{
    char fname[1000];
    int fname_len;
    SYMBOLINX sym[2000];
    STOREPTR sym_caddr[2000];
    int i, sym_count;
    FILE *llpo;
    char buf[BUFSIZ];
    ULONG x;
    char name0[BUFSIZ];
    int arity0;
    STOREPTR cbot, caddr;

    strcpy(fname, fn);
    fname_len = strlen(fname);
    if (fname_len <= 5 || strcmp((fname+fname_len-5), ".llpo") != 0) {
	strcat(fname, ".llpo");
    }
    if ((llpo = fopen(fname, "r")) == NULL) {
	printf("file %s does not exist\n", fname);
	return(1);
    }
    printf("Loading File %s ", fname);
    sym_count = 0;
    load_base = code_bot;
    cbot = code_bot;
    while (! feof(llpo)) {
	if (fgets(buf, BUFSIZ, llpo) == NULL)
	  break;
	if (buf[0] == '\0')
	  continue;
	if (cbot >= heap_top) {
	    printf("CODE area (%ld) overflow\n", CODESIZE);
	    fclose(llpo);
	    return(-1);
	}
	switch (buf[0]) {
	  case '$':	/* symbol definition */
	    buf[strlen(buf)-1] = '\0';
	    strcpy(name0, buf+1);
	    if (fgets(buf, BUFSIZ, llpo) == NULL) {
		printf("LLPAM object file %s format error\n", fname);
		fclose(llpo);
		return(2);
	    }
	    sscanf(buf, "%d %lx", &arity0, &x);
	    if (x == -1L)
	      sym_caddr[sym_count] = PRED_UNDEF;
	    else
	      sym_caddr[sym_count] = load_base + x;
	    sym[sym_count] = am_make_symbol(name0, arity0);
	    sym_count++;
	    break;
	  case '*':	/* absolute */
	    sscanf(buf+1, "%lx", &x);
	    STORE(cbot++) = (CELL)x;
	    break;
	  case '+':	/* relative */
	    sscanf(buf+1, "%lx", &x);
	    STORE(cbot++) = MAKECELL(TAG(x), PTR2INX(load_base + DATA(x)));
	    break;
	  case '@':	/* symbol reference */
	    sscanf(buf+1, "%lx", &x);
	    STORE(cbot++) = MAKECON(sym[CON2SYM(x)]);
	    break;
	  default:
	    printf("LLPAM object file %s format error\n", fname);
	    fclose(llpo);
	    return(2);
	}
    }
    fclose(llpo);
    for (i = 0; i < sym_count; i++) {
	if (sym_caddr[i] != PRED_UNDEF)
	  am_define_pred(sym[i], sym_caddr[i], level);
    }
    code_bot = cbot;
    printf("(%ld words, %ld symbols)\n", code_bot-load_base, sym_count);
    return(0);
}


ULONG name_hash(char *name)
{
    /* See A. V. Aho, R. Sethi, and J. D. Ullman:
       "Compilers Principles, Techniques, and Tools",
       Addison-Wesley, 1990
    */
    ULONG h;

    h = 0;
    for (; *name != '\0'; name++) {
	h = 65599 * h + *name;
    }
    return h;
}


SYMBOLINX am_make_symbol(char *name0, int arity0)
{
    SYMBOLINX s;
    char *name1;

    if (strcmp(name0, "nil") == 0)
      name0 = "[]";
    for (s = 0; s < symbol_bot; s++) {
	if( strcmp(SYMBOL(s).print_name, name0) == 0 &&
	    SYMBOL(s).arity == arity0 ) {
	    return s;
	}
    }
    name1 = (char *)malloc(strlen(name0)+1);
    strcpy(name1, name0);
    s = symbol_bot;
    if (s >= SYMSIZE) {
	printf("SYMBOL area (max %ld symbols) overflow\n", SYMSIZE);
	am_abort(101);
    }
    SYMBOL(s).print_name = name1;
    SYMBOL(s).arity = arity0;
    SYMBOL(s).hash_value = name_hash(name1) + 65599 * arity0;
    SYMBOL(s).codeaddr = PRED_UNDEF;
    SYMBOL(s).plevel = 0;
    SYMBOL(s).res = UNDEF;
    SYMBOL(s).res2 = UNDEF;
    SYMBOL(s).res3 = UNDEF;
    SYMBOL(s).res4 = UNDEF;
    symbol_bot++;
    return s;
}


void am_define_pred(SYMBOLINX sym, STOREPTR caddr, int level)
{
    if (caddr == PRED_UNDEF)
      return;
    if (SYMBOL(sym).codeaddr != PRED_UNDEF) {
	printf("WARNING: Predicate %s/%d is redefined.\n",
	       SYMBOL(sym).print_name, SYMBOL(sym).arity);
    }
    SYMBOL(sym).codeaddr = caddr;
    SYMBOL(sym).plevel = level;
}


/*
 Local Variables:
 mode: c
 c-indent-level: 4
 End:
*/
