/****************************************************************
	LLPAM Emulator
	(read/write token)
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University
****************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include "sysdep.h"
#include "token.h"

#define issolo(c)	((c)=='!' || (c)==';')
#define YES	(1)
#define NO	(0)

/*
   int read_token(char *string, FILE *stream)
     reads 1 token from the stream,
     sets the string, and returns the token type.

   Type		String
   -2		"error message"
   -1		"end_of_file"
   '.'		"."		full stop
   ' '		" "		space or comment or unknown chars
   ','		","
   '('		"("
   ')'		")"
   '['		"["
   ']'		"]"
   '{'		"{"
   '}'		"}"
   ','		","
   '|'		"|"
   'I'		"decimal"	positive integer
   'A'		"atom name"
   'V'		"variable name"
   'S'		"string"

   Floating point numbers and integers with radix are not supported.
*/
int read_token(char *string, FILE *stream)
{
    int c, c1;
    int rc;

    c = fgetc(stream);
    if (c == EOF) {
	strcpy(string, "end_of_file");
	return -1;
    }
    if (isdigit(c)) {
	rc = read_number(c, string, stream);
	if (rc > 0)
	  rc = 'I';
	return rc;
    }
    if (islower(c)) {
	rc = read_word(c, string, stream);
	if (rc > 0)
	  rc = 'A';
	return rc;
    }
    if (isupper(c) || c == '_') {
	rc = read_word(c, string, stream);
	if (rc > 0)
	  rc = 'V';
	return rc;
    }
    switch (c) {
      case '(':
      case ')':
      case '[':
      case ']':
      case '{':
      case '}':
      case ',':
      case '|':
	string[0] = c;
	string[1] = '\0';
	return c;
      case '.':		/* full stop or symbol */
	strcpy(string, ".");
	c1 = fgetc(stream);
	if (c1 == EOF || c1 <= ' ') {
	    return '.';
	}
	ungetc(c1, stream);
	break;
      case '%':		/* one line comment */
	strcpy(string, " ");
	while ((c1 = fgetc(stream)) != '\n') {
	    if (c1 == EOF)
	      return ' ';
	}
	return ' ';
      case '/':		/* start of comment or symbol */
	if ((c1 = fgetc(stream)) == '*') {
	    strcpy(string, " ");
	    while (1) {
		while ((c1 = fgetc(stream)) != '*') {
		    if (c1 == EOF) {
			strcpy(string, "Unexpected end_of_file");
			return -2;
		    }
		}
		if ((c1 = fgetc(stream)) == '/')
		  return ' ';
		ungetc(c1, stream);
	    }
	}
	ungetc(c1, stream);
	break;
      case '\'':
	rc = read_quoted(c, string, stream);
	if (rc > 0)
	  rc = 'A';
	return rc;
      case '"':
	rc = read_quoted(c, string, stream);
	if (rc > 0)
	  rc = 'S';
	return rc;
      default:
	break;
    }
    if (issolo(c)) {
	string[0] = c;
	string[1] = '\0';
	return 'A';
    }
    if (issymbol(c)) {
	rc = read_symbol(c, string, stream);
	if (rc > 0)
	  rc = 'A';
	return rc;
    }
    strcpy(string, " ");
    return ' ';
}


int read_number(int c, char *string, FILE *stream)
{
    int c1;
    char *p;

    ungetc(c, stream);
    for (p = string; ; p++) {
	c1 = fgetc(stream);
	if (! isdigit(c1))
	  break;
	*p = c1;
    }
    *p = '\0';
    ungetc(c1, stream);
    return 1;
}

int read_word(int c, char *string, FILE *stream)
{
    int c1;
    char *p;

    ungetc(c, stream);
    for (p = string; ; p++) {
	c1 = fgetc(stream);
	if (! isalnum(c1) && c1 != '_')
	  break;
	*p = c1;
    }
    *p = '\0';
    ungetc(c1, stream);
    return 1;
}


int read_quoted(int quote, char *string, FILE *stream)
{
    int c1;
    char *p;

    for (p = string; ; p++) {
	c1 = fgetc(stream);
	if (c1 == EOF || c1 == '\n') {
	    *p = '\0';
	    ungetc(c1, stream);
	    strcpy(string, "Unexpected end_of_line");
	    return -2;
	}
	if (c1 == quote) {
	    c1 = fgetc(stream);
	    if (c1 != quote) {
		ungetc(c1, stream);
		*p = '\0';
		return 1;
	    }
	}
	*p = c1;
    }
}


int read_symbol(int c, char *string, FILE *stream)
{
    int c1;
    char *p;

    ungetc(c, stream);
    for (p = string; ; p++) {
	c1 = fgetc(stream);
	if (! issymbol(c1))
	  break;
	*p = c1;
    }
    *p = '\0';
    ungetc(c1, stream);
    return 1;
}


/*
   void write_string(char *string, FILE *stream)
*/
void write_string(char *string, FILE *stream)
{
    fprintf(stream, "%s", string);
}


/*
   void writeq_string(char *string, FILE *stream)
*/
void writeq_string(char *string, FILE *stream)
{
    int needs_quote;
    char *p;

    needs_quote = (string_type(string) == 3);
    if (needs_quote) {
	fprintf(stream, "\'");
	for (p = string; *p; p++) {
	    fprintf(stream, "%c", *p);
	    if (*p == '\'')
	      fprintf(stream, "\'");
	}
	fprintf(stream, "\'");
    } else {
	write_string(string, stream);
    }
}


/*
   int string_type(char *string) returns
	0 : if string is a lower case alphnum
	1 : if string is a symbol
	2 : if string is a solo
	3 : others
*/
int string_type(char *string)
{
    char *p;

    p = string;
    if (strcmp(p, "[]") == 0 || strcmp(p, "{}") == 0 )
      return 0;
    if (*p == '\0' || strcmp(p, ".") == 0)
      return 3;
    if (issolo(*p) && *(p+1) == '\0') 
      return 2;
    if (islower(*p)) {
	for (; *p; p++) {
	    if (! isalnum(*p) && *p != '_')
	      return 3;
	}
	return 0;
    }
    if (issymbol(*p)) {
	for (; *p; p++) {
	    if (! issymbol(*p))
	      return 3;
	}
	return 1;
    }
    return 3;
}


int issymbol(int c)
{
    switch (c) {
      case '+':
      case '-':
      case '*':
      case '/':
      case '\\':
      case '^':
      case '<':
      case '>':
      case '=':
      case '`':
      case '~':
      case ':':
      case '.':
      case '?':
      case '@':
      case '#':
      case '$':
      case '&':
	return 1;
      default:
	return 0;
    }
}

