/**
 * @(#)PrologServer.java
 * @author Naoyuki Tamura (tamura@kobe-u.ac.jp)
 * Dept. CS, Kobe University
 *
 * This program starts a process of Prolog and communicates with it
 * through I/O pipes.
 */
import java.io.*;
import java.util.*;

public class PrologServer implements Runnable {
    private String command;
    private Thread serverThread = null;
    private Process process = null;
    private DataOutputStream out = null;
    //private DataInputStream in = null;
    private BufferedReader in = null;
    private DataInputStream err = null;
    private IOException ioException = null;
    private String message = null;
    private int request;
    private final int NOREQ = 0;
    private final int SENDREQ = 1;
    private final int RECVREQ = 2;

    public PrologServer(String command) throws IOException {
	String cmd[] = new String[3];
	String msg;

	this.command = command;
	System.err.println("Starting process " + command);
	cmd[0] = "/bin/sh";
	cmd[1] = "-c";
	cmd[2] = command;
	process = Runtime.getRuntime().exec(cmd);
	out = new DataOutputStream(process.getOutputStream());
	//in = new DataInputStream(process.getInputStream());
	in = new BufferedReader(
		   new InputStreamReader(process.getInputStream()));
	err = new DataInputStream(process.getErrorStream());
	// wait JAVA READY message
	msg = waitMessage();
	if (! msg.equals("READY")) {
	    close();
	    throw new IOException(command + " can not start");
	}
	// start communication thread
	request = NOREQ;
	serverThread = new Thread(this);
	serverThread.start();
    }

    public void finalize() {
	close();
    }

    public synchronized void close() {
//	System.err.println("*** close");
	if (serverThread != null) {
	    serverThread.stop();
	    serverThread = null;
	    request = NOREQ;
	    ioException = new IOException(command + " is aborted");
	    notifyAll();
	}
	if (process != null) {
	    process.destroy();
	    process = null;
	    System.err.println("End of process " + command);
	}
	if (out != null) {
	    try { out.close(); } catch (IOException e) {}
	    out = null;
	}
	if (in != null) {
	    try { in.close(); } catch (IOException e) {}
	    in = null;
	}
	if (err != null) {
	    try { err.close(); } catch (IOException e) {}
	    err = null;
	}
    }

    private String recv() throws IOException {
	String msg = null;

	if (process == null) {
	    close();
	    throw new IOException(command + " is not working");
	}
	while (err.available() > 0) {
	    // echo error output
	    System.err.write(err.read());
	}
	try {
	    msg = in.readLine();
	} catch (IOException e) {}
	if (msg == null) {
	    close();
	    throw new IOException(command + " has been terminated");
	}
//	System.err.println(msg);
	return msg;
    }

    private String waitMessage() throws IOException {
	String msg;

	do {
	    if (in == null)
		return null;
	    msg = recv();
	} while (! msg.startsWith("JAVA "));
	msg = msg.substring(5);
	return msg;
    }

    private void sendMessage(String message) throws IOException {
	String msg;

//	System.err.println("*** send ");
	msg = waitMessage();
	if (! msg.equals("RECV")) {
	    close();
	    throw new IOException(command + " send protocol error");
	}
	if (out == null) {
	    throw new IOException(command + " send error");
	}
	try {
	    out.writeBytes(message);
	    out.write('\n');
	    out.flush();
//	    System.err.println(message);
	} catch (IOException e) {
	    close();
	    throw new IOException(command + " send error");
	}
    }

    private String receiveMessage() throws IOException {
	String msg;

//	System.err.println("*** receive");
	msg = waitMessage();
	if (! msg.startsWith("SEND ")) {
	    close();
	    throw new IOException(command + " receive protocol error");
	}
	msg = msg.substring(5);
	return msg;
    }

    public synchronized void send(String msg) throws IOException {
	while (request != NOREQ) {
	    try {
		wait();
	    } catch (InterruptedException e) {}
	}
	request = SENDREQ;
	message = msg;
	notifyAll();
	while (request != NOREQ) {
	    try {
		wait();
	    } catch (InterruptedException e) {}
	}
	if (ioException != null) {
	    throw ioException;
	}
    }

    public synchronized String receive() throws IOException {
	while (request != NOREQ) {
	    try {
		wait();
	    } catch (InterruptedException e) {}
	}
	request = RECVREQ;
	notifyAll();
	while (request != NOREQ) {
	    try {
		wait();
	    } catch (InterruptedException e) {}
	}
	if (ioException != null) {
	    throw ioException;
	}
	return message;
    }

    private synchronized void waitRequest() {
	while (request == NOREQ) {
	    try {
		wait();
	    } catch (InterruptedException e) {}
	}
    }

    private synchronized void done() {
	request = NOREQ;
	notifyAll();
    }

    public void run() {
	while (true) {
	    waitRequest();
	    ioException = null;
	    if (request == SENDREQ) {
		try {
		    sendMessage(message);
		} catch (IOException e) {
		    ioException = e;
		}
	    } else if (request == RECVREQ) {
		try {
		    message = receiveMessage();
		} catch (IOException e) {
		    ioException = e;
		}
	    }
	    done();
	}
    }


}

